@@ -7,21 +7,24 @@ use Masterkey::Admin::Record::Layer;
 
 sub new {
     my $class = shift();
-    my($torus, $x, $using_XML_LibXML_Simple) = @_;
+    my($torus, $isWorld, $x, $using_XML_LibXML_Simple) = @_;
 
+    my @list;
     my $this = bless {
 	torus => $torus,
+	isWorld => $isWorld,
 	uri => delete $x->{URI},
 	type => delete $x->{type},
-	layers => {},
+	layerlist => \@list,	# Main list of layers, may have duplicates names
+	layermap => {},		# Hash into list, good enough for most purposes
     }, $class;
 
     Scalar::Util::weaken($this->{torus});
     foreach my $y (@{ $x->{layer} }) {
 	my $name = $y->{name};
-	$this->{layers}->{_default} = $this->{layers}->{$name} = 
-	    new Masterkey::Admin::Record::Layer($this, $y,
-						$using_XML_LibXML_Simple);
+	my $layer = new Masterkey::Admin::Record::Layer($this, $y, $using_XML_LibXML_Simple);
+	push @list, [ $name, $layer ];
+	$this->{layermap}->{$name} = $layer;
     }
 
     if (0) {
@@ -42,8 +45,8 @@ sub new {
 }
 
 sub torus { shift()->{torus} }
+sub isWorldRecord { shift()->{isWorld} }
 sub type { shift()->{type} }
-sub layer { shift()->{layers}->{_default}->layer() }
 sub uri { 
     my $this = shift();
     my($newval) = @_;
@@ -51,13 +54,6 @@ sub uri {
     return $this->{uri}
 }
 
-sub isWorldRecord {
-	my $this = shift();
-	#warn "*", $this->displayName(), " layer = '", $this->layer(), "' from {", join(",", sort keys %{$this->{layers}}), "}";
-	return (!exists $this->{layers}->{override} &&
-		!exists $this->{layers}->{'master-override'});
-}
-
 
 # fields() and field() both delegate to a nominated layer, if one is
 # specified, or to the default layer if none is.
@@ -66,12 +62,12 @@ sub fields {
     my $this = shift();
     my($layer) = @_;
 
-    $layer = "_default" if !defined $layer;
-    my $l = $this->{layers}->{$layer};
+    $layer = $this->_bestLayer(0) if !defined $layer;
+    my $l = $this->{layermap}->{$layer};
     return $l->fields() if defined $l;
 
     warn "$this->fields('$layer') has no layer: use one of (" .
-	join(", ", map { "'$_'" } sort keys %{ $this->{layers} }) . ")";
+	join(", ", map { "'$_'" } sort keys %{ $this->{layermap} }) . ")";
     return ();
 }
 
@@ -79,13 +75,13 @@ sub field {
     my $this = shift();
     my($name, $lname) = @_;
 
-    if (!defined $lname && exists $this->{layers}->{'master-override'}) {
+    if (!defined $lname &&
+	exists $this->{layermap}->{'master-override'} &&
+	!exists $this->{layermap}->{'override'}) {
 	# Special case: master records have only one layer
 	$lname = 'master-override';
-	my $x = $this->{layers}->{'master-override'}->{fields};
-	warn "using master-override layer for field '$name': {" .
-	    join(", ", map { "$_='" . $x->{$_} . "'" } sort keys %$x) .
-	    "}";
+	my $x = $this->{layermap}->{'master-override'}->{fields};
+	#warn "using master-override layer for field '$name': {" . join(", ", map { "$_='" . $x->{$_} . "'" } sort keys %$x) . "}";
     }
 
     if (!defined $lname && $name eq "id") {
@@ -93,7 +89,7 @@ sub field {
 	# has no override layer), be sure to return the ID from the
 	# "original" layer rather than the different one in the
 	# "final" layer. See bug MKII-1724.
-	$lname = "original" if !exists $this->{layers}->{override};
+	$lname = "original" if $this->isWorldRecord();
     }
 
     if (!defined $lname) {
@@ -106,24 +102,56 @@ sub field {
 		$this->field($name, "original"));
     }
 
-    my $layer = $this->{layers}->{$lname};
+    my $layer = $this->{layermap}->{$lname};
     return undef if !defined $layer;
     #die "layer '$lname' not defined" if !defined $layer;
     return $layer->field($name);
 }
 
+# Returns all layers with the specified name
+# Only really useful for walking through multiple master-overrides.
+#
+sub layers {
+    my $this = shift();
+    my($lname) = @_;
+
+    my @res;
+    foreach my $ref (@{ $this->{layerlist} }) {
+	my($name, $layer) = @$ref;
+	push @res, $layer if $name eq $lname;
+    }
+
+    return @res;
+}
+
+
 sub setField {
     my $this = shift();
     my($name, $value) = @_;
 
-    # Currently only needed for master records, so to avoid surprises
-    # we reject the call if this is not a master.
-    
-    die "setField() called on non-Master record"
-	if !exists $this->{layers}->{'master-override'};
+    my $lname = $this->_bestLayer(1);
+    die "in setField: candidate layers all missing"
+	if !defined $lname;
 
     ### We should use the Layer API instead of poking inside hashes.
-    $this->{layers}->{'master-override'}->{fields}->{$name} = $value;
+    my $layer = $this->{layermap}->{$lname};
+    $layer->{fields}->{$name} = $value;
+}
+
+sub _bestLayer {
+    my $this = shift();
+    my($forWriting) = @_;
+
+    my @lnames = $forWriting ?
+	qw(override master-override final) :
+	qw(final override original master-override);
+
+    foreach my $lname (@lnames) {
+	if (defined $this->{layermap}->{$lname}) {
+	    return $lname;
+	}
+    }
+    return undef;
 }
 
 sub _show {
@@ -141,21 +169,6 @@ sub displayName {
     my $nval = $this->field("displayName", "override");
     my $oval = $this->field("displayName", "original");
 
-
-    ### Debugging for a specific problem -- will be removed later
-    if (0 && ((defined $oval && $oval =~ /Free/) ||
-	      (defined $nval && $nval =~ /Free/))) {
-	my($package, $filename, $line) = caller(1);
-	warn("displayName(oval" . (defined $oval ? "='$oval'" : " undefined"),
-	     ", nval" . (defined $nval ? "='$nval'" : " undefined"), ")",
-	     ", caller=$filename:$line\n");
-	use Data::Dumper;
-	local $Data::Dumper::Indent = 1;
-	local $this->{torus} = undef;
-	warn Dumper($this);
-    }
-
-
     if (defined $nval && defined $oval) {
 	return "$nval ($oval)";
     } elsif (defined $nval) {
@@ -212,7 +225,7 @@ sub update {
     die "$this->update() called on a non-auth record"
 	if !$this->torus()->{isAuth};
 
-    $this->torus()->update(undef,
+    $this->torus()->update(0, undef,
 			   $this->field("id"),
 			   $this->type(),
 			   {},
@@ -223,7 +236,7 @@ sub delete {
     my $this = shift();
     my $torus = $this->torus();
 
-    $this->torus()->delete($this->field("realm"), $this->field("id"));
+    $this->torus()->deleteByArgs(realm => $this->field("realm"), id => $this->field("id"));
 }
 
 sub render {
