@@ -10,19 +10,23 @@ use URI::Escape qw(uri_escape_utf8);
 use Compress::Zlib;
 use XML::LibXML;
 use Carp qw(confess);
+use Encode;
+use Time::HiRes qw/gettimeofday tv_interval/;
+
 use Masterkey::Admin::Record;
 use Masterkey::Admin::ResultSet;
 use Masterkey::Admin::Parent;
 use Masterkey::Admin::Master;
 use Masterkey::Admin::Properties;
 
-
 our $using_XML_LibXML_Simple;
 
 BEGIN {
     # Use XML::LibXML::Simple when possible, but it's not available on
     # all platforms (e.g. MacOS X v10.5.8) so be prepared to back down
-    # to XML::Simple.
+    # to XML::Simple. Pathetically, the API of XML::LibXML::Simple is
+    # not backwards compatible, so we need to know which we're
+    # using. Sheesh.
     eval {
 	require 'XML/LibXML/Simple.pm';
 	import XML::LibXML::Simple;
@@ -40,7 +44,7 @@ sub new {
     my $class = shift();
     my($admin, $ws) = @_;
 
-    warn "WARNING: making $class with non-torus2 WS '$ws'"
+    die "WARNING: making $class with non-torus2 WS '$ws'"
 	if $ws !~ /\/torus2\//;
 
     my $this = bless {
@@ -76,15 +80,14 @@ sub log {
 
 sub _logTorusOp {
     my $this = shift();
-    my($op, @params) = @_;
+    my($op, %args) = @_;
 
     my $ws = $this->ws();
     $ws =~ s/^.*?\/(torus2?|realm)\///;
     my $s = "";
-    while (@params) {
-	my $name = shift @params;
-	my $val = shift @params;
-	next if !defined $val;
+    foreach my $name (sort keys %args) {
+	my $val = $args{$name};
+	next if $name eq 'listType' || !defined $val;
 	$s .= ", " if $s ne "";
 	$val = $val->field("identityId") if $name eq "user";
 	$s .= "$name='$val'";
@@ -96,55 +99,84 @@ sub _logTorusOp {
 # Official entry points, including logging
 sub records {
     my $this = shift();
-    return $this->_list("records", @_);
+    my($realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive) = @_;
+
+    return $this->_listByArgs(listType => "records",
+			      realm => $realm,
+			      query => $query,
+			      user => $user,
+			      noConstraint => $noConstraint,
+			      facets => $facets,
+			      start => $start,
+			      count => $count,
+			      recursive => $recursive);
 }
 
 sub world {
     my $this = shift();
     my($realm, $query, $facets, $start, $count, $recursive) = @_;
-    return $this->_list("world", $realm, $query, undef, undef, $facets, $start, $count, $recursive);
+
+    return $this->_listByArgs(listType => "world",
+			      realm => $realm,
+			      query => $query,
+			      facets => $facets,
+			      start => $start,
+			      count => $count,
+			      recursive => $recursive);
 }
 
 sub merged {
     my $this = shift();
-    return $this->_list("merged", @_);
+    my($realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive) = @_;
+
+    return $this->_listByArgs(listType => "merged",
+			      realm => $realm,
+			      query => $query,
+			      user => $user,
+			      noConstraint => $noConstraint,
+			      facets => $facets,
+			      start => $start,
+			      count => $count,
+			      recursive => $recursive);
 }
 
 
 # Internal entry point, includes logging
-sub _list {
+sub _listByArgs {
     my $this = shift();
-    my($listType, $realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive) = @_;
+    my(%args) = @_;
 
+    my $query = $args{query};
     if (ref $query eq "HASH") {
-	$query = _hash2cql(%$query);
+	$args{query} = _hash2cql(%$query);
     }
 
-    $this->_logTorusOp($listType, realm=>$realm, query=>$query,
-		       user=>$user, noConstraint=>$noConstraint,
-		       facets=>$facets, start=>$start, count=>$count,
-		       recursive=>$recursive);
-    # Can't just use @_ here, as $query may have been rewritten
-    return $this->_list_nolog($listType, $realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive);
+    $this->_logTorusOp($args{listType}, %args);
+    return $this->_listNologByArgs(%args);
 }
 
 
 # Internal entry point without logging
-sub _list_nolog {
+sub _listNologByArgs {
     my $this = shift();
-    my($listType, $realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive) = @_;
+    my(%args) = @_;
 
-    my $uri = $this->_listURI($listType, $realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive);
+    my $uri = $this->_listURIbyArgs(%args);
     my $simple = $this->_getAsXML($uri);
 
+    my $listType = $args{listType};
     my $ntotal = $simple->{total};
     my $nlocal = $simple->{local};
     my $noverrides = $simple->{overriden};
     my $nretrieved = $simple->{count};
+
     die "no count" if !defined $nretrieved;
+    my $count = $args{count};
     die "impossible count $nretrieved>$count"
 	if defined $count && $nretrieved > $count;
-    $start = 0 if !defined $start;
+
+    my $start = $args{start};
+    $start = 0 if !defined $start || $start eq '';
     my $start2 = $simple->{start};
     die "impossible start $start2 != $start"
 	if defined $start2 && $start2 != $start;
@@ -153,19 +185,36 @@ sub _list_nolog {
     my $baseUri = $uri;
     $baseUri =~ s/\?.*//;
     foreach my $x (@{ $simple->{record} }) {
-	my $rec = new Masterkey::Admin::Record($this, $x, $using_XML_LibXML_Simple);
+	my $isWorld;
+	if ($listType eq 'world') {
+	    $isWorld = 1;
+	} elsif ($listType eq 'records') {
+	    $isWorld = 0;
+	} elsif ($listType eq 'merged') {
+	    my $haveOverride = 0;
+	    foreach my $layer (@{ $x->{layer} }) {
+		$haveOverride = 1 if $layer->{name} eq 'override';
+	    }
+	    $isWorld = !$haveOverride;
+	} else {
+	    die "unexpected listType '$listType'";
+	}
+
+	my $rec = new Masterkey::Admin::Record($this, $isWorld, $x, $using_XML_LibXML_Simple);
 	if (!$rec->uri()) {
 	    # Newer Toruses don't set this, so fill it in by hand.
 	    # This gluing process is pretty ugly.  ### Note, too, that
 	    # it DOES NOT WORK for records in /merged/, which must
 	    # instead be fetched from either /records/ or /world/
-	    $rec->uri($baseUri . $rec->field("id") . "/");
+	    my $id = $rec->field("id");
+	    die "id undefined" if !defined $id;
+	    $rec->uri($baseUri . $id . "/");
 	}
 	push @records, $rec;
     }
 
     my $group;
-    if (defined $facets) {
+    if (defined $args{facets}) {
 	use Masterkey::Admin::Facets;
 	$group = Masterkey::Admin::Facets->group_from_xml(
 	    max => 5, uri => "", xml => $simple->{facets}->[0]->{facet});
@@ -177,18 +226,13 @@ sub _list_nolog {
 }
 
 
-sub recordsURI {
+sub recordsURIbyArgs {
     my $this = shift();
-    my($realm, $query, $user, $noConstraint) = @_;
+    my(%args) = @_;
+
+    my $realm = $args{realm};
     die "listURI invoked on 'world'" if defined $realm && $realm eq "world";
-    return $this->_listURI("records", $realm, $query, $user, $noConstraint);
-}
-
-sub worldURI {
-    my $this = shift();
-    my($realm, $query) = @_;
-
-    return $this->_listURI("world", $realm, $query);
+    return $this->_listURIbyArgs(listType => "records", %args);
 }
 
 
@@ -196,9 +240,28 @@ sub worldURI {
 # constraint, in which case it supplies values to be substituted into
 # the constraint query.  That's it.  Otherwise, $user is unused.
 #
-sub _listURI {
+sub _listURIbyArgs {
     my $this = shift();
-    my($listType, $realm, $query, $user, $noConstraint, $facets, $start, $count, $recursive) = @_;
+    my(%args) = @_;
+
+    if (0) {
+	foreach my $key (sort keys %args) {
+	    delete $args{$key} if !defined $args{$key};
+	}
+	warn("args = { " , join(", ", map { "$_='" . $args{$_} . "'" } sort keys %args) , " }");
+    }
+
+    my $listType = $args{listType};
+    my $realm = $args{realm};
+    my $layers = $args{layers};
+    my $query = $args{query};
+    my $user = $args{user};
+    my $noConstraint = $args{noConstraint};
+    my $facets = $args{facets};
+    my $start = $args{start};
+    my $count = $args{count};
+    my $recursive = $args{recursive};
+
     $query = undef if defined $query && $query eq "";
     if (!$noConstraint && !$this->{isAuth} &&
 	$this->admin()->conf("constrainIdentity")) {
@@ -209,7 +272,7 @@ sub _listURI {
     $uri .= "$listType/";
 
     my %params = ();
-    $params{layers} = "override,original,final";
+    $params{layers} = $layers if defined $layers;
     $params{query} = $query if defined $query;
     $params{facets} = $facets if defined $facets;
     $params{start} = $start if defined $start;
@@ -222,7 +285,7 @@ sub _listURI {
 			   sort keys %params);
     }
 
-    #warn "_listURI(listType=", _render($listType), " realm=", _render($realm), ", query=", _render($query), ", user=", _render($user), ", noConstraint=", _render($noConstraint), ") = '$uri'\n";
+    #warn "_listURIbyArgs(listType=", _render($listType), " realm=", _render($realm), ", query=", _render($query), ", user=", _render($user), ", noConstraint=", _render($noConstraint), ") = '$uri'\n";
     return $uri;
 }
 
@@ -250,27 +313,27 @@ sub _torusURI {
 }
 
 
-sub record {
+sub recordByArgs {
     my $this = shift();
-    my($realm, $id) = @_;
+    my(%args) = @_;
 
-    $this->_logTorusOp("record", realm=>$realm, id=>$id);
-    return $this->_record("records", $realm, $id);
+    $this->_logTorusOp("record", %args);
+    return $this->_recordByArgs(listType => "records", %args);
 }
 
-sub worldRecord {
+sub worldRecordByArgs {
     my $this = shift();
-    my($realm, $worldId) = @_;
+    my(%args) = @_;
 
-    $this->_logTorusOp("worldRecord", realm=>$realm, worldId=>$worldId);
-    return $this->_record("world", $realm, $worldId);
+    $this->_logTorusOp("worldRecord", %args);
+    return $this->_recordByArgs(listType => "world", %args);
 }
 
-sub _record {
+sub _recordByArgs {
     my $this = shift();
-    my($listType, $realm, $id) = @_;
+    my(%args) = @_;
 
-    my $url = $this->recordURI($listType, $realm, $id);
+    my $url = $this->recordURIbyArgs(%args);
     my $simple = $this->_getAsXML($url);
     # Defend against subtle consequences of incorrect URIs.
     die "record() retrieved a list" if $simple->{record};
@@ -278,17 +341,24 @@ sub _record {
     # Newer Toruses don't set this, so fill it in by hand
     $url =~ s/\?.*//;
     $simple->{URI} ||= $url;
-    return new Masterkey::Admin::Record($this, $simple,
+
+    return new Masterkey::Admin::Record($this, $args{listType} eq 'world', $simple,
 					$using_XML_LibXML_Simple);
 }
 
-### Trailing slash currently needed due to Torus bug #4135
-sub recordURI {
-    my $this = shift();
-    my($listType, $realm, $id) = @_;
 
-    confess "recordURI invoked on 'world'" if defined $realm && $realm eq "world";
-    my $listURI = $this->_listURI($listType, $realm, undef, undef, 1);
+sub recordURIbyArgs {
+    my $this = shift();
+    my(%args) = @_;
+
+    # Safety check for an old "can't happen" that once happened
+    my $realm = $args{realm};
+    confess "recordURIbyArgs invoked on 'world'" if defined $realm && $realm eq "world";
+
+    my $listURI = $this->_listURIbyArgs(%args, noConstraint => 1);
+
+    ### Considing making _listURIbyArgs() do the following steps with $args{$id} if present
+    my $id = $args{id};
     if ($listURI =~ s/\/\?/\/$id\/?/) {
 	return $listURI;
     }
@@ -301,16 +371,34 @@ sub _getAsXML {
     my($uri) = @_;
 
     my $res = $this->_get($uri);
-    my $xml = $res->content()
+    my $xml = $res->decoded_content('default_charset' => 'UTF-8')
 	or die "no content for '$uri'";
 
     return XMLin($xml, ForceArray => 1, KeyAttr => []);
 }
 
+sub time_diff {
+    my $this = shift;
+    my $t0 = shift || $this->{'_time_start'};
+
+    # time im milli seconds
+    my $elapsed = int(tv_interval($t0) * 1000);
+
+    return $elapsed;
+}
+
+sub time_start {
+    my $this = shift;
+
+    return $this->{'_time_start'} = [gettimeofday];
+}
+
 sub _get {
     my $this = shift();
     my($uri) = @_;
 
+    $this->time_start;
+
     if ($uri =~ /^file:/) {
 	# Hack needed for when we make fake Toruses in the filesystem,
 	# since we can't have a file and a directory of the same name.
@@ -318,9 +406,13 @@ sub _get {
 	$uri .= ".xml";
     }
 
-    $this->log(Masterkey::Admin::LogLevel::TORUS, "GET $uri");
+
     my $res = $this->{ua}->get($uri, "Accept-Encoding" => "gzip")
 	or die "no response for '$uri'\n";
+
+    my $elapsed = $this->time_diff;
+    $this->log(Masterkey::Admin::LogLevel::TORUS, "GET $uri [$elapsed ms]");
+
     $this->_decompress_content($res); ### Why doesn't LWP do this itself?
     if (!$res->is_success()) {
 	$this->_http_fatal("GET", $uri, $res);
@@ -340,41 +432,6 @@ sub _get {
     return $res;
 }
 
-sub _put {
-    my $this = shift();
-    my($uri, $xml) = @_;
-
-    $this->log(Masterkey::Admin::LogLevel::TORUS, "PUT $uri");
-    $this->log(Masterkey::Admin::LogLevel::WSPUT, $xml);
-    my $res = $this->{ua}->request(PUT($uri, Content => $xml,
-		   "Content-Type" => "application/xml"))
-	or die "no response for PUT to '$uri'\n";
-
-    if (!$res->is_success()) {
-	$this->_http_fatal("PUT to", $uri, $res);
-    }
-
-    return $res;
-}
-
-sub _post {
-    my $this = shift();
-    my($uri, $xml) = @_;
-
-    $this->log(Masterkey::Admin::LogLevel::TORUS, "POST $uri");
-    $this->log(Masterkey::Admin::LogLevel::WSPOST, $xml);
-    my $res = $this->{ua}->request(POST($uri, Content => $xml,
-		   "Content-Type" => "application/xml"))
-	or die "no response for POST to '$uri'\n";
-
-    if (!$res->is_success()) {
-	$this->_http_fatal("POST to", $uri, $res);
-    }
-
-    return $res;
-}
-
-
 sub _decompress_content {
     my $this = shift();
     my($res) = @_;
@@ -387,24 +444,75 @@ sub _decompress_content {
 	$res->content($content);
 	$res->header("content-encoding" => "");
     	$this->log(Masterkey::Admin::LogLevel::COMPRESS,
-		   "decompressed $oldlength to ", length($content), " chars ",
+		   "decompressed $oldlength to ", length($content), " bytes ",
 		   "(", int(100 * $oldlength / length($content)), "%)");
     } else {
     	$this->log(Masterkey::Admin::LogLevel::COMPRESS,
-		   "read ", length($res->content()), " chars, not compressed");
+		   "read ", length($res->content()), " bytes, not compressed");
     }
 }
 
+# convert unicode characters to bytes before sending
+# it over the wire. Use it for all data exchanges with
+# external APIs!
+sub unicode2octets {
+    my $data = shift;
+
+    return Encode::encode( "utf-8", $data, Encode::FB_DEFAULT );
+}
+
+sub _put {
+    my $this = shift();
+    return $this->_put_or_post('PUT', Masterkey::Admin::LogLevel::WSPUT, \&PUT, @_);
+}
+
+sub _post {
+    my $this = shift();
+    return $this->_put_or_post('POST', Masterkey::Admin::LogLevel::WSPOST, \&POST, @_);
+}
+
+sub _put_or_post {
+    my $this = shift();
+    my($label, $loglevel, $method, $uri, $xml) = @_;
+
+    $this->log($loglevel, "data is utf8: " . Encode::is_utf8($xml));
+    if (!Encode::is_utf8($xml)) {
+    	$this->log($loglevel, "XML is not utf8, set utf8 flag");
+    	Encode::_utf8_on($xml) if !Encode::is_utf8($xml);
+    }
+    $xml = &unicode2octets($xml);
+
+    $this->log($loglevel, $xml);
+    $this->time_start;
+    my $res = $this->{ua}->request($method->($uri, Content => $xml,
+		   "Content-Type" => "application/xml"))
+	or die "no response for $label to '$uri'\n";
+
+    my $elapsed = $this->time_diff;
+    $this->log(Masterkey::Admin::LogLevel::TORUS, "x$label $uri [$elapsed ms]");
+
+    if (!$res->is_success()) {
+	$this->_http_fatal("$label to", $uri, $res);
+    }
+
+    return $res;
+}
+
 
 # $goneHash is a reference to a hash whose keys are the names of
 # fields that should be removed completely from the record.
 sub update {
     my $this = shift();
-    my($realm, $id, $type, $goneHash, %data) = @_;
+    my($master, $realm, $id, $type, $goneHash, %data) = @_;
 
-    $this->_logTorusOp("update", realm=>$realm, id=>$id, type=>$type);
-    my $uri = $this->recordURI("records", $realm, $id);
-    my $xml = $this->_makeXML($type, $goneHash, %data);
+    $this->_logTorusOp("update", master=>$master, realm=>$realm, id=>$id, type=>$type);
+    my $uri;
+    if ($master) {
+	$uri = $this->masterURI($realm, $id);
+    } else {
+	$uri = $this->recordURIbyArgs(listType => "records", realm => $realm, id => $id);
+    }
+    my $xml = $this->_makeXML($master, $type, $goneHash, %data);
     $this->_put($uri, $xml);
 }
 
@@ -437,11 +545,10 @@ sub add {
     my($uri, $xml);
     if ($master) {
 	$uri = $this->mastersURI($realm);
-	$xml = _master_xml(undef, "XXX", "cql.allRecords=1", {}, %data);
     } else {
-	$uri = $this->recordsURI($realm, undef, $user, 1);
-	$xml = $this->_makeXML(undef, {}, %data, %extra);
+	$uri = $this->recordsURIbyArgs(realm => $realm, user => $user, noConstraint => 1);
     }
+    $xml = $this->_makeXML($master, undef, {}, %data, %extra);
     my $res = $this->_post($uri, $xml);
 
     # Content is now empty, but the Location header has something like:
@@ -457,35 +564,32 @@ sub add {
     return $id;
 }
 
-sub delete {
+sub deleteByArgs {
     my $this = shift();
-    my($realm, $id) = @_;
+    my(%args) = @_;
 
-    $this->_logTorusOp("delete", realm=>$realm, id=>$id);
-
-    my $uri = $this->recordURI("records", $realm, $id);
+    $this->_logTorusOp("delete", %args);
+    my $uri = $this->recordURIbyArgs(listType => "records", %args);
     $this->_delete_by_uri($uri);
 }
 
-sub disable {
+sub disableByArgs {
     my $this = shift();
-    my($realm, $id) = @_;
+    my(%args) = @_;
 
-    $this->_logTorusOp("disable", realm=>$realm, id=>$id);
-
-    my $uri = $this->recordURI("records", $realm, $id);
-    my $xml = $this->_makeXML(undef, {}, disabled => "yes");
+    $this->_logTorusOp("disable", %args);
+    my $uri = $this->recordURIbyArgs(listType => "records", %args);
+    my $xml = $this->_makeXML(0, undef, {}, disabled => "yes");
     $this->_put($uri, $xml);
 }
 
-sub enable {
+sub enableByArgs {
     my $this = shift();
-    my($realm, $id) = @_;
+    my(%args) = @_;
 
-    $this->_logTorusOp("enable", realm=>$realm, id=>$id);
-
-    my $uri = $this->recordURI("records", $realm, $id);
-    my $xml = $this->_makeXML(undef, {}, disabled => "no");
+    $this->_logTorusOp("enable", %args);
+    my $uri = $this->recordURIbyArgs(listType => "records", %args);
+    my $xml = $this->_makeXML(0, undef, {}, disabled => "no");
     $this->_put($uri, $xml);
 }
 
@@ -522,7 +626,7 @@ sub scan {
     $this->_logTorusOp("scan", realm=>$realm, query=>$query,
 		      user=>$user, field=>$field);
 
-    my $rs = $this->_list_nolog("records", $realm, $query, $user);
+    my $rs = $this->_listNologByArgs(listType => "records", realm => $realm, query => $query, user => $user);
     my $records = $rs->records();
     my %register;
     foreach my $record (@$records) {
@@ -539,7 +643,7 @@ sub user1 {
 
     if (@_ == 2 && $_[0] eq "id") {
 	# Special-case a nicer query for the common lookup by ID
-	return $this->record(undef, $_[1]);
+	return $this->recordByArgs(realm => undef, id => $_[1]);
     }
 
     my $rs = $this->records(undef, { @_ });
@@ -576,7 +680,7 @@ sub _check_unique {
 	qq[$name == "$val"];
     } @conds);
 
-    my $rs = $this->_list_nolog("records", $realm, $cql, $user);
+    my $rs = $this->_listNologByArgs(listType => "records", realm => $realm, query => $cql, user => $user);
     my $records = $rs->records();
     return if !@$records;
 
@@ -592,7 +696,7 @@ sub _hash2cql {
 	$val =~ s/[""*?]/\\$&/g;
 	qq[$_=="$val"];
     } sort keys %args);
-    
+
     return $query;
 }
 
@@ -601,7 +705,7 @@ sub _http_fatal {
     my($method, $uri, $res) = @_;
 
     my $message = "can't $method '$uri': " . $res->message();
-    my $content = $res->content();
+    my $content = $res->decoded_content('default_charset' => 'UTF-8');
     $message .= " ($content)" if defined $content && $content ne "";
     $message .= "\n";
     warn $message; # For the HTTP server log
@@ -610,7 +714,7 @@ sub _http_fatal {
 
 sub _makeXML {
     my $this = shift();
-    my($type, $goneHash, %data) = @_;
+    my($master, $type, $goneHash, %data) = @_;
 
     $type = $this->admin()->conf('type')
 	if !defined $type;
@@ -622,10 +726,16 @@ sub _makeXML {
     $root->setAttribute(type => $type);
 
     my $layer = $doc->createElement('layer');
-    $layer->setAttribute(name => 'override');
+    $layer->setAttribute(name => $master ? 'master-override' : 'override');
     $root->appendChild($layer);
 
     for my $name (sort keys %allKeys) {
+	if ($name =~ s/^@//) {
+	    # Set attribute instead of field
+	    $layer->setAttribute($name => $data{'@' . $name});
+	    next;
+	}
+
 	my $tag = $doc->createElement($name);
 	if (exists $goneHash->{$name}) {
 	    $tag->setAttribute(gone => 'yes');
@@ -831,7 +941,7 @@ sub master {
     $this->_logTorusOp("master", realm=>$realm, id=>$id);
     my $uri = $this->masterURI($realm, $id);
     my $simple = $this->_getAsXML($uri);
-    return new Masterkey::Admin::Master($this, $simple);
+    return new Masterkey::Admin::Master($this, $simple, $using_XML_LibXML_Simple);
 }
 
 
@@ -919,22 +1029,6 @@ sub parentsURI {
     return $uri;
 }
 
-### The "master" code here is disturbingly similar to the "parent" code.
-
-### This is never used
-sub updateMaster {
-    my $this = shift();
-    my($realm, $id, $label, $matchQuery, $goneHash, %data) = @_;
-
-    $this->_logTorusOp("updateMaster", realm=>$realm, id=>$id, label=>$label,
-		      matchQuery=>$matchQuery);
-
-    my $uri = $this->masterURI($realm, $id);
-    my $xml = _master_xml($id, $label, $matchQuery, %data);
-    $this->_put($uri, $xml);
-}
-
-
 ### disturbingly similar to _makeXML()
 sub _master_xml {
     my($id, $label, $matchQuery, $goneHash, %data) = @_;
